/* File:      cov_mat.C
   Date:      7/9/02 (modified from code written in fall 2001)
   Author:    C. Paciorek
   Purpose:   code for creating, factoring and multiplying pos. def. matrices
   Compile:   g++ -c -I/usr/statlocal/vni/CTT3.0/include
   Usage:     
   Arguments: 
   Requires:  cov_mat.h, base.h, locations.h, corr.h, imsl.h, permtools.h
   Reference: George & Liu 1981: Computer solution of large sparse positive definite systems
              the sparse matrix structure and manipulation follows the code in this reference, 
	      as do the algorithms for factoring and multiplying the matrices;
              remember that IMSL matrices are row-major 

              kernel approach follows that of Dave Higdon of Los Alamos Nat'l Labs in 
              several papers; the specific use of a small number of basis kernels is in his
              Env Ecol Stats 5:173-190, 1998
*/

#include "base.h"
#include "cov_mat.h"
#include "locations.h"
#include "corr.h"
#include "imsl.h"
#include "permtools.h"

sparse_cov_mat::sparse_cov_mat(int size_in){
  size=size_in;
  num_dim=2;
}

sparse_cov_mat_stat::sparse_cov_mat_stat(int size_in,double* corr_parm_in) : sparse_cov_mat(size_in){
  corr_parm=corr_parm_in;  
  sym_chol();              // symbolic factorization -- sparse structure already created in locations.C code
  fill(FIRST);
  gen_chol(WARN);              // do numerical factorization
  update(ACCEPT);
}

sparse_cov_mat_nonstat_basis::sparse_cov_mat_nonstat_basis(int size_in,double* variances_in, double* noise_var_in,double* nu_in,double* kappa1_in, double* kappa2_in,double** kern_eigenval1_in,double** kern_eigenval2_in,double** kern_eigenvec1_in): sparse_cov_mat(size_in){

  kern_eigenval1=kern_eigenval1_in;
  kern_eigenvec1=kern_eigenvec1_in;
  kern_eigenval2=kern_eigenval2_in;

  variances=variances_in;
  noise_var=noise_var_in;
  nu=nu_in;
  kappa1=kappa1_in;

  basis_kernels=new mat_pp<double>(locations->num_kernels,num_dim*num_dim);
  kernels=new mat_pp<double>(size,num_dim*num_dim);
  summat=new double[num_dim*num_dim];
  vec=new double[num_dim];

  sym_chol();                 
  /* done in dataset::datacov_change
  fill(DATA_COVAR_EIGVEC,FIRST);
  gen_chol(WARN);              // do numerical factorization
  update(ACCEPT);
  */
}

sparse_cov_mat_nonstat::sparse_cov_mat_nonstat(int size_in,double* variances_in, double* noise_var_in,double* eigenval1_in,double* eigenval2_in,double* eigenvec1_in,double* eigenvec2_in): sparse_cov_mat(size_in){

  eigenval_mat=new double[num_dim*num_dim];
  eigenval1=eigenval1_in;
  eigenvec1=eigenvec1_in;
  eigenval2=eigenval2_in;
  eigenvec2=eigenvec2_in;
  variances=variances_in;
  noise_var=noise_var_in;

  kernel=new mat_pp<double>(size,num_dim*num_dim);
  eigenvecs=new mat_pp<double>(size,num_dim*num_dim);
  eigenvecs_stored=new mat_pp<double>(size,num_dim*num_dim);
  summat=new double[num_dim*num_dim];
  vec=new double[num_dim];
  chol=new double[num_dim*num_dim];
  tmp=new double[num_dim*num_dim];

  sym_chol();                 
  /* done in dataset
  fill(DATA_COVAR_EIGVEC,FIRST);
  gen_chol(WARN);              // do numerical factorization
  update(ACCEPT);
  */
}

sparse_cov_mat_fullstat::sparse_cov_mat_fullstat(int size_in,double* variances_in, double* noise_var_in,double* nu_in,double* corr_parm_in): sparse_cov_mat(size_in){

  corr_parm=corr_parm_in;
  variances=variances_in;
  noise_var=noise_var_in;
  nu=nu_in;

  sym_chol();                 
  /*  done in dataset
      fill(DATA_KAPPA,FIRST);
  gen_chol(WARN);              // do numerical factorization
  update(ACCEPT);
  */
}

sparse_cov_mat_fixed::sparse_cov_mat_fixed(int size_in,double* variances_in, double* noise_var_in,string file_in): sparse_cov_mat(size_in){

  variances=variances_in;
  noise_var=noise_var_in;

  sym_chol();                 

  corrmat_in=new mat_pp<double>(size_in,size_in);
  corrmat_in->read(file_in,BY_ROW,MATRIX);

  /*  done in dataset
      fill(DATA_KAPPA,FIRST);
  gen_chol(WARN);              // do numerical factorization
  update(ACCEPT);
  */
}

sparse_cov_mat::~sparse_cov_mat(void){
  delete [] nzsub;
  delete [] xlnz;
  delete [] xnzsub;
  delete [] lnz_L;
  delete [] lnz_C;
  delete [] diag_L;
  delete [] diag_C;
  delete [] lnz_L_stored; 
  delete [] diag_L_stored;
  delete [] lnz_C_stored; 
  delete [] diag_C_stored;
  delete distances;
}

sparse_cov_mat_stat::~sparse_cov_mat_stat(void){
  delete [] lnz_Csum;
  delete [] lnz_L_Csum;
  delete [] diag_Csum;
  delete [] diag_L_Csum;
}

sparse_cov_mat_nonstat_basis::~sparse_cov_mat_nonstat_basis(void){ 
  delete kernels;
  delete basis_kernels;
  delete [] summat;
  delete [] vec;
  delete [] lnz_R;
  delete [] diag_R;
  delete [] lnz_R_stored;
  delete [] diag_R_stored;
  delete [] Euc_u;
  delete [] Euc_v;
}

sparse_cov_mat_nonstat::~sparse_cov_mat_nonstat(void){ 
  delete [] eigenval_mat;
  delete kernel;
  delete eigenvecs;
  delete eigenvecs_stored;
  delete [] summat;
  delete [] tmp;
  delete [] vec;
  delete [] chol;
  delete [] lnz_R;
  delete [] diag_R;
  delete [] lnz_R_stored;
  delete [] diag_R_stored;
  delete [] Euc_u;
  delete [] Euc_v;
}

sparse_cov_mat_fullstat::~sparse_cov_mat_fullstat(void){ 
  delete [] lnz_R;
  delete [] diag_R;
  delete [] lnz_R_stored;
  delete [] diag_R_stored;
}

sparse_cov_mat_fixed::~sparse_cov_mat_fixed(void){ 
  delete [] lnz_R;
  delete [] diag_R;
  delete corrmat_in;
}

void sparse_cov_mat_stat::add_info(double* data_diag_C_in,double* data_lnz_C_in,double* sigma_in,double divisor_in,double* etahat_in,double* delta_in){
  data_diag_C=data_diag_C_in;
  data_lnz_C=data_lnz_C_in;
  divisor=divisor_in;
  sigma=sigma_in;
  etahat=etahat_in;
  delta=delta_in;
}

void sparse_cov_mat_stat::fill(int first){
  // fill stationary CORRELATION matrix
  int index, col, i, istart, istop, ii, isub;
  if(first==FIRST){
    for(index=0;index<size;index++){
      diag_C[index]=1.0;
    }
    index=0;

    // initialize sparse structure -- see details above
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1]-1;
	distances[index]=locations->ang_distance(isub,col,RADIANS,PERMUTE);
	lnz_C[index]=fillij(index);
	i++; index++;
      }
    }
  }
  else{
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1]-1;
	lnz_C[index]=fillij(index);
	i++; index++;
      }
    }
  }
}

double sparse_cov_mat_stat::fillij(int index){
  // calculates proposed corr values based on proposed correlation parameter value
  return(matern_corr_ang(distances[index],(*corr_parm),smoothness));
}

void sparse_cov_mat_nonstat_basis::fill(int which,int first){
  // fills individual values of sparse structure of covar matrix

  int i,k, index,istart,istop,ii,col,isub, p;
  double c,s, sumwgt,wgt;
  
  if(first==FIRST){
    // fill kernels
    for(k=0;k<locations->num_kernels;k++){  
      c=cos(*kern_eigenvec1[k]);
      s=sin(*kern_eigenvec1[k]);
      // calculate kernel from eigen-decomposition
      basis_kernels->vals[k][0]=*kern_eigenval1[k]*c*c+*kern_eigenval2[k]*s*s;
      basis_kernels->vals[k][1]=(*kern_eigenval1[k]-*kern_eigenval2[k])*c*s;
      basis_kernels->vals[k][2]=basis_kernels->vals[k][1];
      basis_kernels->vals[k][3]=*kern_eigenval1[k]*s*s+*kern_eigenval2[k]*c*c;
    }
    // find kernels at each location
    for(i=0;i<size;i++){  
      // k=0
      wgt=exp(-(pow(locations->locn_kernel_dists->vals[i][0]/(*kappa1),2.0)));
      for(p=0;p<num_dim*num_dim;p++){
        kernels->vals[i][p]=wgt*basis_kernels->vals[0][p];
      }
      sumwgt=wgt;
      for(k=1;k<locations->num_kernels;k++){  
        wgt=exp(-(pow(locations->locn_kernel_dists->vals[i][k]/(*kappa1),2.0)));
        for(p=0;p<num_dim*num_dim;p++){
          kernels->vals[i][p]+=wgt*basis_kernels->vals[k][p];
        }
        sumwgt+=wgt;
      }
      for(p=0;p<num_dim*num_dim;p++){
        kernels->vals[i][p]/=sumwgt;
      }
    }
    
    // ofstream tmpout("/TOIL-U2/paciorek/research/code/spat17/covinfo.dat");
    

    for(i=0;i<size;i++){
      diag_R[i]=1.0;                        // working with base correlation matrix and vector of variances
      diag_C[i]=variances[i]+(*noise_var);
    }

    
    // fill off-diags in sparse structure -- see details above
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
        isub=nzsub[i-1]-1;
        // calculate angular distance between the pair of locations and store it
        //distances[index]=locations->ang_distance(isub,col,RADIANS,PERMUTE);
        // calculate Euclidean u and v distances between the pair of locations and store it

        /*
      if((isub==2 && col==133) || (isub==133 && col==2)){
        cout << "weird\n";
          }
        */

        locations->euc_distance(isub,col,&(Euc_u[index]),&(Euc_v[index]),PERMUTE);
        // calc proposed off-diag value of corr matrix
        lnz_R[index]=fillij(isub,col,Euc_u[index],Euc_v[index]);
        // lnz_R[index]=fillij(locations->perm[isub]-1,locations->perm[col]-1,distances[index],Euc_u[index],Euc_v[index]);  -- not sure why I was doing the permutation here
        // calc proposed off-diag value of covariance matrix
        lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
        //cout << index << " " << isub << " " << col << " " << Euc_u[index] << " " << Euc_v[index] << " " << lnz_R[index] << " " << locations->perm[isub] << " " << locations->perm[col] << endl; 

        //tmpout << lnz_R[index] << " ";
               
        i++; index++;
      }
    }
    /*
    tmpout << endl;
    tmpout.close();
    */
  }
  else{
    if(which==DELTA){               // variance of indep. error
      for(i=0;i<size;i++){          // replace old value with proposed value
	diag_C[i]=variances[i]+(*noise_var);
      }
    }
    if(which==ETA){
      for(i=0;i<size;i++){                    // calc proposed diag values
	diag_C[i]=variances[i]+(*noise_var);
      }
      index=0;                                // start with first non-zero location
      for(col=0;col<size-1;col++){            // step through columns, no off-diag in last column
	istart=xlnz[col];                     // xlnz holds indices for lnz
	istop=xlnz[col+1]-1;                  // i.e. istart and istop are arrayindex+1
	i=xnzsub[col];                        // xnzsub holds indices for nzsub, which contains row values
	for(ii=istart;ii<=istop;ii++){        // only deal with values in current column
	  isub=nzsub[i-1]-1;                  // isub is the current row
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);  // calc proposed off-diags
	  i++; index++;                       // step through all non-zero values
	}     
      }
    }
    if(which==DATA_EIG_KAPPA || which==DATA_COVAR_EIGVEC || which==DATA_COVAR_EIGVAL || which==NU){
      // calculate kernel matrices
      for(k=0;k<locations->num_kernels;k++){  
        c=cos(*kern_eigenvec1[k]);
        s=sin(*kern_eigenvec1[k]);
        basis_kernels->vals[k][0]=*kern_eigenval1[k]*c*c+*kern_eigenval2[k]*s*s;
        basis_kernels->vals[k][1]=(*kern_eigenval1[k]-*kern_eigenval2[k])*c*s;
        basis_kernels->vals[k][2]=basis_kernels->vals[k][1];
        basis_kernels->vals[k][3]=*kern_eigenval1[k]*s*s+*kern_eigenval2[k]*c*c;
      }
      // done filling kernels
      // find kernels at each location
      for(i=0;i<size;i++){  
        // k=0
        wgt=exp(-(pow(locations->locn_kernel_dists->vals[i][0]/(*kappa1),2.0)));
        for(p=0;p<num_dim*num_dim;p++){
          kernels->vals[i][p]=wgt*basis_kernels->vals[0][p];
        }
        sumwgt=wgt;
        for(k=1;k<locations->num_kernels;k++){  
          wgt=exp(-(pow(locations->locn_kernel_dists->vals[i][k]/(*kappa1),2.0)));
          for(p=0;p<num_dim*num_dim;p++){
            kernels->vals[i][p]+=wgt*basis_kernels->vals[k][p];
          }
          sumwgt+=wgt;
        }
        for(p=0;p<num_dim*num_dim;p++){
          kernels->vals[i][p]/=sumwgt;
        }
      }
      
      // finished calculating kernel matrices
 /*  the following code walks through the sparse structure of the matrix by row within column
      istart and istop determine how many non-zero values are in each column
      index walks through all the non-zero values one by one
      xnzsub tells where in nzsub the row values for the current column start
      nzsub contains the row values for every non-zero element
      values within xnsub and nzsub are 1,..., hence the -1's sprinkled throughout to
      get C style 0,... indexing
      note that some 'non-zero' values might be zero
      I may want to consider changing isub to row, nzsub to row_vals and xnzsub to row_vals_pointer
  */
      index=0;
      for(col=0;col<size-1;col++){  
	istart=xlnz[col];  
	istop=xlnz[col+1]-1;
	i=xnzsub[col];
	for(ii=istart;ii<=istop;ii++){
	  isub=nzsub[i-1]-1;
	  lnz_R[index]=fillij(isub,col,Euc_u[index],Euc_v[index]);
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	  i++; index++;
	}
      }
    }
  }
}

double sparse_cov_mat_nonstat_basis::fillij(int i, int j,double u, double v){
  // this calculates Higdon-extended Matern kernel correlation between locations i and j, where u and v are the Euclidean distance between the locations, i and j are the permuted versions
  //double deti=1.0; double detj=1.0; double detsum=1.0;
  double deti; double detj; double detsum=0.0;
  double Q;
  int p;
  double outval=0.0;
  //  double distvec[2];
  //distvec[0]=u;
  //distvec[1]=v;
  
  deti=kernels->vals[i][0]*kernels->vals[i][3]-kernels->vals[i][1]*kernels->vals[i][1];
  detj=kernels->vals[j][0]*kernels->vals[j][3]-kernels->vals[j][1]*kernels->vals[j][1];
  for(p=0;p<num_dim*num_dim;p++){
    summat[p]=kernels->vals[i][p]+kernels->vals[j][p];
  }
  
  /* new code */
  detsum=summat[0]*summat[3]-summat[1]*summat[1];
  Q=(2.0/detsum)*(u*u*summat[3]-2.0*u*v*summat[1]+v*v*summat[0]);  // this is (u v)((1/2)*(sig_x+sig_y))^-1(u v)
  outval=(2.0*pow(deti*detj,0.25)*matern_corr_euc(sqrt(Q),1.0,*nu))/(pow(detsum,0.5));
  return(outval);

}

/* not currently used */
void sparse_cov_mat_nonstat::fill(int which,int first){
  // fills individual values of sparse structure of covar matrix

  // if I use this for 2-d problem, write out code explicitly and remove IMSL calls for 2d matrix manipulations

  int i, index,istart,istop,ii,col,isub, p;
  double len;
    /*  the following code walks through the sparse structure of the matrix by row within column
	istart and istop determine how many non-zero values are in each column
	index walks through all the non-zero values one by one
	xnzsub tells where in nzsub the row values for the current column start
	nzsub contains the row values for every non-zero element
	values within xnsub and nzsub are 1,..., hence the -1's sprinkled throughout to
	get C style 0,... indexing
	note that some 'non-zero' values might be zero
	I may want to consider changing isub to row, nzsub to row_vals and xnzsub to row_vals_pointer
     */

  if(first==FIRST){   // initialize values
    for(p=0;p<num_dim*num_dim;p++){
      eigenval_mat[p]=0.0;
    }
    // fill kernels
    for(i=0;i<size;i++){  
      len=pow(eigenvec1[i]*eigenvec1[i]+eigenvec2[i]*eigenvec2[i],0.5);
      eigenvecs->vals[i][0]=eigenvec1[i]/len;
      eigenvecs->vals[i][2]=eigenvec2[i]/len;
      eigenvecs->vals[i][1]=-eigenvecs->vals[i][2];
      eigenvecs->vals[i][3]=eigenvecs->vals[i][0];
    }
    for(i=0;i<size;i++){ 
      eigenval_mat[0]=eigenval1[i];
      eigenval_mat[3]=eigenval2[i];
      imsl_d_mat_mul_rect("A*B",IMSL_A_MATRIX,num_dim,num_dim,eigenvecs->vals[i],IMSL_B_MATRIX,num_dim,num_dim,eigenval_mat,IMSL_RETURN_USER,tmp,0);
      imsl_d_mat_mul_rect("A*trans(B)",IMSL_A_MATRIX,num_dim,num_dim,tmp,IMSL_B_MATRIX,num_dim,num_dim,eigenvecs->vals[i],IMSL_RETURN_USER,kernel->vals[i],0);
      if(kernel->vals[i][0]<=0.0 || kernel->vals[i][3]<=0.0){
	cout << "warning: neg variances\n"; cout.flush();
      }
      if(fabs(kernel->vals[i][1])>=sqrt(kernel->vals[i][0]*kernel->vals[i][3])){
	cout << "warning: not pos def\n"; cout.flush();
      }
    }
    // done filling kernels
    for(i=0;i<size;i++){
      diag_R[i]=1.0;                        // working with base correlation matrix and vector of variances
      diag_C[i]=variances[i]+(*noise_var);
    }

    // fill off-diags in sparse structure -- see details above
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1]-1;
	// calculate angular distance between the pair of locations and store it
	//distances[index]=locations->ang_distance(isub,col,RADIANS,PERMUTE);
	// calculate Euclidean u and v distances between the pair of locations and store it
	locations->euc_distance(isub,col,&(Euc_u[index]),&(Euc_v[index]),PERMUTE);
	// calc proposed off-diag value of corr matrix
	lnz_R[index]=fillij(isub,col,Euc_u[index],Euc_v[index]);
	// calc proposed off-diag value of covariance matrix
	lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	i++; index++;
      }
    }
  }
  else{
    if(which==DELTA){               // variance of indep. error
      for(i=0;i<size;i++){          // replace old value with proposed value
	diag_C[i]= variances[i]+(*noise_var);
      }
    }
    if(which==ETA){
      for(i=0;i<size;i++){                    // calc proposed diag values
	diag_C[i]=variances[i]+(*noise_var);
      }
      index=0;                                // start with first non-zero location
      for(col=0;col<size-1;col++){            // step through columns, no off-diag in last column
	istart=xlnz[col];                     // xlnz holds indices for lnz
	istop=xlnz[col+1]-1;                  // i.e. istart and istop are arrayindex+1
	i=xnzsub[col];                        // xnzsub holds indices for nzsub, which contains row values
	for(ii=istart;ii<=istop;ii++){        // only deal with values in current column
	  isub=nzsub[i-1]-1;                  // isub is the current row
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);  // calc proposed off-diags
	  i++; index++;                       // step through all non-zero values
	}     
      }
    }
    if(which==DATA_COVAR_EIGVEC || which==DATA_COVAR_EIGVAL){
      // calculate kernel matrices
      if(which==DATA_COVAR_EIGVEC){
	for(i=0;i<size;i++){  
	  len=pow(eigenvec1[i]*eigenvec1[i]+eigenvec2[i]*eigenvec2[i],0.5);
	  eigenvecs->vals[i][0]=eigenvec1[i]/len;
	  eigenvecs->vals[i][2]=eigenvec2[i]/len;
	  eigenvecs->vals[i][1]=-eigenvecs->vals[i][2];
	  eigenvecs->vals[i][3]=eigenvecs->vals[i][0];
	}
      }
      for(i=0;i<size;i++){ 
	eigenval_mat[0]=eigenval1[i];
	eigenval_mat[3]=eigenval2[i];
	imsl_d_mat_mul_rect("A*B",IMSL_A_MATRIX,num_dim,num_dim,eigenvecs->vals[i],IMSL_B_MATRIX,num_dim,num_dim,eigenval_mat,IMSL_RETURN_USER,tmp,0);
	imsl_d_mat_mul_rect("A*trans(B)",IMSL_A_MATRIX,num_dim,num_dim,tmp,IMSL_B_MATRIX,num_dim,num_dim,eigenvecs->vals[i],IMSL_RETURN_USER,kernel->vals[i],0);
	if(kernel->vals[i][0]<=0.0 || kernel->vals[i][3]<=0.0){
	  cout << "warning: neg variances\n"; cout.flush();
	}
	if(fabs(kernel->vals[i][1])>=sqrt(kernel->vals[i][0]*kernel->vals[i][3])){
	  cout << "warning: not pos def\n"; cout.flush();
	}
      }
      // finished calculating kernel matrices
      index=0;
      for(col=0;col<size-1;col++){  
	istart=xlnz[col];  
	istop=xlnz[col+1]-1;
	i=xnzsub[col];
	for(ii=istart;ii<=istop;ii++){
	  isub=nzsub[i-1]-1;
	  lnz_R[index]=fillij(isub,col,Euc_u[index],Euc_v[index]);
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	  i++; index++;
	}
      }
    }
  }
}

// note: this still uses squ exponential form of covariance
double sparse_cov_mat_nonstat::fillij(int i, int j,double u, double v){
  // this calculates Higdon kernel correlation between locations i and j, where u and v are the Euclidean distance between the locations, i and j are the permuted versions
  double sum=0.0;
  //double deti=1.0; double detj=1.0; double detsum=1.0;
  double deti; double detj; double detsum=0.0;
  int p;
  double outval=0.0;
  double distvec[2];
  distvec[0]=u;
  distvec[1]=v;

  deti=log(eigenval1[i])+log(eigenval2[i]);
  detj=log(eigenval1[j])+log(eigenval2[j]);

  for(p=0;p<num_dim*num_dim;p++){
    summat[p]=kernel->vals[i][p]+kernel->vals[j][p];
  }
  imsl_d_lin_sol_posdef(num_dim,summat,distvec,IMSL_RETURN_USER,vec,IMSL_FACTOR_USER,chol,0);
  for(p=0;p<num_dim;p++){
    sum+=vec[p]*distvec[p];
  }
  for(p=0;p<num_dim;p++){
    detsum+=log(chol[p*num_dim+p]);
  }
  //return(pow(2,num_dim/2)*pow(deti*detj,0.25)*exp(-0.5*sum)/detsum);  // pow(detsum,0.5)); -- using diag of Chol so already have square root
  outval=log(2)+0.25*deti+0.25*detj-0.5*sum-detsum;
  //no 1/2 factor for detsum becuz it's from the cholesky
  return(exp(outval));
}


void sparse_cov_mat_fullstat::fill(int which,int first){
  // fills individual values of sparse structure of covar matrix

  int i, index,istart,istop,ii,col,isub;

    /*  the following code walks through the sparse structure of the matrix by row within column
	istart and istop determine how many non-zero values are in each column
	index walks through all the non-zero values one by one
	xnzsub tells where in nzsub the row values for the current column start
	nzsub contains the row values for every non-zero element
	values within xnsub and nzsub are 1,..., hence the -1's sprinkled throughout to
	get C style 0,... indexing
	note that some 'non-zero' values might be zero
	I may want to consider changing isub to row, nzsub to row_vals and xnzsub to row_vals_pointer
     */

  if(first==FIRST){   // initialize values
    for(i=0;i<size;i++){
      diag_R[i]=1.0;                        // working with base correlation matrix and vector of variances
      diag_C[i]=variances[i]+(*noise_var);
    }

    // fill off-diags in sparse structure -- see details above
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1]-1;
	// calculate angular distance between the pair of locations and store it
	distances[index]=locations->ang_distance(isub,col,RADIANS,PERMUTE);
	// calculate Euclidean u and v distances between the pair of locations and store it
	// calc proposed off-diag value of corr matrix
	lnz_R[index]=fillij(index);
	lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	i++; index++;
      }
    }
  }
  else{
    if(which==DELTA){               // variance of indep. error
      for(i=0;i<size;i++){          // replace old value with proposed value
	diag_C[i]= variances[i]+(*noise_var);
      }
    }
    if(which==ETA){
      for(i=0;i<size;i++){                    // calc proposed diag values
	diag_C[i]=variances[i]+(*noise_var);
      }
      index=0;                                // start with first non-zero location
      for(col=0;col<size-1;col++){            // step through columns, no off-diag in last column
	istart=xlnz[col];                     // xlnz holds indices for lnz
	istop=xlnz[col+1]-1;                  // i.e. istart and istop are arrayindex+1
	i=xnzsub[col];                        // xnzsub holds indices for nzsub, which contains row values
	for(ii=istart;ii<=istop;ii++){        // only deal with values in current column
	  isub=nzsub[i-1]-1;                  // isub is the current row
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);  // calc proposed off-diags
	  i++; index++;                       // step through all non-zero values
	}     
      }
    }
    if(which==DATA_KAPPA || which==NU){
      // finished calculating kernel matrices
      index=0;
      for(col=0;col<size-1;col++){  
	istart=xlnz[col];  
	istop=xlnz[col+1]-1;
	i=xnzsub[col];
	for(ii=istart;ii<=istop;ii++){
	  isub=nzsub[i-1]-1;
	  lnz_R[index]=fillij(index);
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	  i++; index++;
	}
      }
    }
  }
}

double sparse_cov_mat_fullstat::fillij(int index){
  // calculates proposed corr values based on proposed correlation parameter value
    return(matern_corr_ang(distances[index],(*corr_parm),*nu));
}

void sparse_cov_mat_fixed::fill(int which,int first){
  // fills individual values of sparse structure of covar matrix

  int i, index,istart,istop,ii,col,isub;

    /*  the following code walks through the sparse structure of the matrix by row within column
	istart and istop determine how many non-zero values are in each column
	index walks through all the non-zero values one by one
	xnzsub tells where in nzsub the row values for the current column start
	nzsub contains the row values for every non-zero element
	values within xnsub and nzsub are 1,..., hence the -1's sprinkled throughout to
	get C style 0,... indexing
	note that some 'non-zero' values might be zero
	I may want to consider changing isub to row, nzsub to row_vals and xnzsub to row_vals_pointer
     */

  if(first==FIRST){   // initialize values
    for(i=0;i<size;i++){
      diag_R[i]=1.0;                        // working with base correlation matrix and vector of variances
      diag_C[i]=variances[i]+(*noise_var);
    }

    // fill off-diags in sparse structure -- see details above
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1]-1;
	lnz_R[index]=fillij(isub,col);
	lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);
	i++; index++;
      }
    }
  }
  else{
    if(which==DELTA){               // variance of indep. error
      for(i=0;i<size;i++){          // replace old value with proposed value
	diag_C[i]= variances[i]+(*noise_var);
      }
    }
    if(which==ETA){
      for(i=0;i<size;i++){                    // calc proposed diag values
	diag_C[i]=variances[i]+(*noise_var);
      }
      index=0;                                // start with first non-zero location
      for(col=0;col<size-1;col++){            // step through columns, no off-diag in last column
	istart=xlnz[col];                     // xlnz holds indices for lnz
	istop=xlnz[col+1]-1;                  // i.e. istart and istop are arrayindex+1
	i=xnzsub[col];                        // xnzsub holds indices for nzsub, which contains row values
	for(ii=istart;ii<=istop;ii++){        // only deal with values in current column
	  isub=nzsub[i-1]-1;                  // isub is the current row
	  lnz_C[index]=lnz_R[index]*pow(variances[isub]*variances[col],0.5);  // calc proposed off-diags
	  i++; index++;                       // step through all non-zero values
	}     
      }
    }
  }
}

double sparse_cov_mat_fixed::fillij(int isub,int col){
  // fills individual elements from input correlation matrix
  int newrow=locations->perm[isub]-PERMSTART1;
  int newcol=locations->perm[col]-PERMSTART1;
  return(corrmat_in->vals[newrow][newcol]);
}

void sparse_cov_mat_stat::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  // only need the Cholesky factor values since others are not used in the chain
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    // needed in Csum calculations; possibly elsewhere
    copy(size,diag_C,diag_C_stored);
    copy(maxlnz,lnz_C,lnz_C_stored);
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(size,diag_C_stored,diag_C);
    copy(maxlnz,lnz_C_stored,lnz_C);
  }
}

void sparse_cov_mat_nonstat_basis::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    copy(size,diag_R,diag_R_stored);
    copy(maxlnz,lnz_R,lnz_R_stored);
    // these are needed in Csum calculations
    copy(size,diag_C,diag_C_stored);
    copy(maxlnz,lnz_C,lnz_C_stored);
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(size,diag_R_stored,diag_R);
    copy(maxlnz,lnz_R_stored,lnz_R);
    copy(size,diag_C_stored,diag_C);
    copy(maxlnz,lnz_C_stored,lnz_C);
  }
}

void sparse_cov_mat_nonstat::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  int i;
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    copy(size,diag_R,diag_R_stored);
    copy(maxlnz,lnz_R,lnz_R_stored);
    copy(size,diag_C,diag_C_stored);
    copy(maxlnz,lnz_C,lnz_C_stored);
    for(i=0;i<size;i++){    // save eigenvecs to be used when only eigenval proposed
      copy(num_dim*num_dim,eigenvecs->vals[i],eigenvecs_stored->vals[i]);
    }
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(size,diag_R_stored,diag_R);
    copy(maxlnz,lnz_R_stored,lnz_R);
    copy(size,diag_C_stored,diag_C);
    copy(maxlnz,lnz_C_stored,lnz_C);
    for(i=0;i<size;i++){
      copy(num_dim*num_dim,eigenvecs_stored->vals[i],eigenvecs->vals[i]);
    } 
  }
}

void sparse_cov_mat_fullstat::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    copy(size,diag_R,diag_R_stored);
    copy(maxlnz,lnz_R,lnz_R_stored);
    copy(size,diag_C,diag_C_stored);
    copy(maxlnz,lnz_C,lnz_C_stored);
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(size,diag_R_stored,diag_R);
    copy(maxlnz,lnz_R_stored,lnz_R);
    copy(size,diag_C_stored,diag_C);
    copy(maxlnz,lnz_C_stored,lnz_C);
  }
}

void sparse_cov_mat_fixed::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    // needed in Csum calculations
    copy(size,diag_C,diag_C_stored);
    copy(maxlnz,lnz_C,lnz_C_stored);
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(size,diag_C_stored,diag_C);
    copy(maxlnz,lnz_C_stored,lnz_C);
  }
}

double sparse_cov_mat::det_log_squ_root(void){
  // calculates log square root of determinant as log of product of diagonals of Cholesky factor
  int i;
  double tmp=0.0;
  tmp=0.0;
  for(i=0;i<size;i++){
    if(diag_L[i]<.000000001){
      tmp+=(-27.0);
    } 
    else{
      tmp+=log(diag_L[i]);
    }
  }
  return(tmp);
}
/*
double sparse_cov_mat_fullstat::det_log_squ_root(void){
  // calculates log square root of determinant as log of product of diagonals of Cholesky factor
  gen_chol2(WARN);
  

  int i;
  double tmp=0.0;
  for(i=0;i<size;i++){
    tmp+=0.5*log(variances[i]);
  }
  cout << "Var contrib: " << tmp << endl;

  tmp=0.0;
  for(i=0;i<size;i++){
    tmp+=log(diag_L[i]);
  }

  cout << "R contrib: " << tmp << endl;

  for(i=0;i<size;i++){
    tmp+=0.5*log(variances[i]);
  }

  gen_chol(WARN);
  return(tmp);
}
double sparse_cov_mat_nonstat_basis::det_log_squ_root(void){
  // calculates log square root of determinant as log of product of diagonals of Cholesky factor
  gen_chol2(WARN);
  

  int i;
  double tmp=0.0;
  for(i=0;i<size;i++){
    tmp+=0.5*log(variances[i]);
  }
  cout << "Var contrib: " << tmp << endl;

  tmp=0.0;
  for(i=0;i<size;i++){
    tmp+=log(diag_L[i]);
  }

  cout << "R contrib: " << tmp << endl;

  for(i=0;i<size;i++){
    tmp+=0.5*log(variances[i]);
  }

  gen_chol(WARN);
  return(tmp);
}
*/
void sparse_cov_mat::sym_chol(void){
  // calculate symbolic Cholesky factorization -- i.e., set up sparse structure based on adjacency information
  int* rchlnk; int* mrglnk; int* marker; int* temp_nzsub; 

  rchlnk = new int[size];  // temp storage for use by Fortran routine
  mrglnk=new int[size];    //  "
  marker=new int[size];    //  "
  temp_nzsub=new int[size*size];   // row indices
  xlnz=new int[size+1];            // pointers to values (where each column starts)
  xnzsub=new int[size+1];          // pointers to row indices (where each column starts)
  maxsub = size*size;  // input size of nzsub -- rewritten within smbfct_

  smbfct_(&size,locations->xadj,locations->adjncy,locations->perm,locations->invp,xlnz,&maxlnz,xnzsub,temp_nzsub,&maxsub,rchlnk,mrglnk, marker);  // symbolic factorization call; locations objects already filled; this call fills the sparse structure
 
  nzsub=new int[maxsub];
  copy(maxsub,temp_nzsub,nzsub);
  lnz_L=new double[maxlnz];
  lnz_C=new double[maxlnz];
  diag_L=new double[size];
  diag_C=new double[size];
  lnz_L_stored=new double[maxlnz];
  diag_L_stored=new double[size];
  lnz_C_stored=new double[maxlnz];
  diag_C_stored=new double[size];

  distances=new double[maxlnz];

  delete [] rchlnk;
  delete [] mrglnk;
  delete [] marker;
  delete [] temp_nzsub;

}

void sparse_cov_mat_stat::sym_chol(void){
  sparse_cov_mat::sym_chol();
  lnz_Csum=new double[maxlnz];
  lnz_L_Csum=new double[maxlnz];
  diag_Csum=new double[size];
  diag_L_Csum=new double[size];
}

void sparse_cov_mat_nonstat_basis::sym_chol(void){
  sparse_cov_mat::sym_chol();
  lnz_R=new double[maxlnz];
  diag_R=new double[size];
  lnz_R_stored=new double[maxlnz];
  diag_R_stored=new double[size];
  Euc_u=new double[maxlnz];
  Euc_v=new double[maxlnz];
}

void sparse_cov_mat_nonstat::sym_chol(void){
  sparse_cov_mat::sym_chol();
  lnz_R=new double[maxlnz];
  diag_R=new double[size];
  lnz_R_stored=new double[maxlnz];
  diag_R_stored=new double[size];
  Euc_u=new double[maxlnz];
  Euc_v=new double[maxlnz];
}

void sparse_cov_mat_fullstat::sym_chol(void){
  // nonstationary symbolic factorization is same as stationary, except extra info needs to be stored since stationary matrices are all correlation matrices
  sparse_cov_mat::sym_chol();
  lnz_R=new double[maxlnz];
  diag_R=new double[size];
  lnz_R_stored=new double[maxlnz];
  diag_R_stored=new double[size];
}

void sparse_cov_mat_fixed::sym_chol(void){
  // nonstationary symbolic factorization is same as stationary, except extra info needs to be stored since stationary matrices are all correlation matrices
  sparse_cov_mat::sym_chol();
  lnz_R=new double[maxlnz];
  diag_R=new double[size];
}

void sparse_cov_mat::gen_chol(int warn){
  // calculate numerical Cholesky factorization
  double eps=TOL;
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors

  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_C,temp_lnz_L);
  copy(size,diag_C,temp_diag_L);

  // Fortran routine for calculating numeric factorization
  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);
  
  if(warn>1){
    cout << "Warning, zeroed out column in C matrix Cholesky\n";
  }

  copy(maxlnz,temp_lnz_L,lnz_L);
  copy(size,temp_diag_L,diag_L);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}
/*
void sparse_cov_mat_fullstat::gen_chol2(int warn){
  // calculate numerical Cholesky factorization
  double eps=TOL;
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors

  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_R,temp_lnz_L);
  copy(size,diag_R,temp_diag_L);

  // Fortran routine for calculating numeric factorization

  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);

  copy(maxlnz,temp_lnz_L,lnz_L);
  copy(size,temp_diag_L,diag_L);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}

void sparse_cov_mat_nonstat_basis::gen_chol2(int warn){
  // calculate numerical Cholesky factorization
  double eps=TOL;
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors

  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_R,temp_lnz_L);
  copy(size,diag_R,temp_diag_L);

  // Fortran routine for calculating numeric factorization

  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);

  copy(maxlnz,temp_lnz_L,lnz_L);
  copy(size,temp_diag_L,diag_L);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}
*/
void sparse_cov_mat::solve(double* rhs_in, double* output){
  // calculates C^{-1} dot b where b=rhs_in
  // this follows the code of George & Liu's gsslv, p. 156-7

  // to solve Cb=L*t(L)b=y, let z=t(L)b, solve Lz=y for z; then solve
  // t(L)b=z for b;  i.e. y=C^{-1}b

  double rhsj, s;
  int i, ii, istop, istart,isub,j;

  // changed to double in regr4 for better avoiding underflow
  // using floats is probably not much faster because the computational burden is in the Cholesky not here in solve
  double* rhs=new double[size];
  double* diag=new double[size]; 
  double* lnz=new double[maxlnz];

  copy(maxlnz,lnz_L,lnz);  
  copy(size,diag_L,diag);  
  copy(size,rhs_in,rhs);
  
  // forward substitution -- solve Lz=y
  int flag=0;
  for(j=0;j<size;j++){
    if(diag[j]<TOL){
      rhs[j]=0.0;
      if(flag==0){
	flag=j;
      }
    }
    else{
      rhsj= rhs[j]/diag[j];
      rhs[j]= rhsj;
      //rhsj= (*rhs)(j)/diag[j];
      //(*rhs)(j)= rhsj;
      if(j==size-1 ) {break;}
      istart=xlnz[j];  
      if(j==size-2) {istop=istart;}
      else
	{
	  istop=xlnz[j+1]-1;
	}
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++)
	{
	  isub=nzsub[i-1];
	  rhs[isub-1]=rhs[isub-1]-lnz[ii-1]*rhsj;
	     //(*rhs)(isub-1)=(*rhs)(isub-1)-lnz[ii-1]*rhsj;
	  i++;
	}
    }
  }
  if(flag){
    cout << "Warning: zeroed out column in C matrix chol: " << flag << "\n"; cout.flush();
  } 
  // backward substitution -- solve t(L)b=z


  for(j=size-1;j>=0;j--) {
    if(diag[j]<TOL){
      rhs[j]=0.0;
    }
    else{
      s=rhs[j];
      //s=(*rhs)(j);
      if(j<size-1){
	istart=xlnz[j];
	if(j==size-2) {istop=istart;}
	else{
	  istop=xlnz[j+1]-1;
	}
	i=xnzsub[j];
	for(ii=istart;ii<=istop;ii++){
	  isub=nzsub[i-1];
	  s=s-lnz[ii-1]*rhs[isub-1];
	  //s=s-lnz[ii-1]*(*rhs)(isub-1);
	  i++;
	}
      }
      rhs[j]=s/diag[j];
      //(*rhs)(j)=s/diag[j];
    }
  }
  
  copy(size,rhs,output);

  delete [] rhs;
  delete [] diag;
  delete [] lnz;
}

void sparse_cov_mat::Linvmult(double* rhs_in, double* output){
  // calculates L^{-1} dot b where b=rhs_in
  // this follows the code of George & Liu's gsslv, p. 156-7, except just the 
  // first half of the solve routine, essentially it backs out the output from
  // L*output=rhs_in


  float rhsj;
  int i, ii, istop, istart,isub,j;

  float* rhs=new float[size];
  float* diag=new float[size]; 
  float* lnz=new float[maxlnz];

  copy(maxlnz,lnz_L,lnz);
  copy(size,diag_L,diag);
  copy(size,rhs_in,rhs);
  
  // forward substitution -- solve Lz=y

  for(j=0;j<size;j++){
    if(diag[j]<TOL){
      rhs[j]=0.0;  // this fxn is only used for initialization so should be ok
    }
    else{
      rhsj= rhs[j]/diag[j];
      rhs[j]= rhsj;
      //rhsj= (*rhs)(j)/diag[j];
      //(*rhs)(j)= rhsj;
      if(j==size-1 ) {break;}
      istart=xlnz[j];  
      if(j==size-2) {istop=istart;}
      else{
	  istop=xlnz[j+1]-1;
	}
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++)
	{
	  isub=nzsub[i-1];
	  rhs[isub-1]=rhs[isub-1]-lnz[ii-1]*rhsj;
	     //(*rhs)(isub-1)=(*rhs)(isub-1)-lnz[ii-1]*rhsj;
	  i++;
	}
    }
  }
  copy(size,rhs,output);

  delete [] rhs;
  delete [] diag;
  delete [] lnz;
}

void sparse_cov_mat::Lcheck(double* newvec,double* oldvec){
  // for elements whose col is zeroed out, change newvec to old vec
  for(int j=0;j<size;j++){
    if(diag_L[j]<TOL){
      newvec[j]=oldvec[j];
    }
  }
}

/* code for creating full matrix from columnwise representation:
set up mat_pp or 2-dim matrix
then run through with code below:
  for(j=0;j<size-1;j++){
      lnz2[j][j]=diag_L[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	lnz2[row][j]=lnz_L[ii-1];
	i++;
      }
    }
  }
*/

void sparse_cov_mat::Lmult(double* b,double* output)
{
  // calculates L*b

  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
      output[j]=0.0;
    }
  // j is col
  for(j=0;j<size-1;j++){
    if(diag_L[j]>=TOL){
      output[j]+=diag_L[j]*b[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	output[row]+=lnz_L[ii-1]*b[j];
	i++;
      }
    }
  }
  output[size-1]+=diag_L[size-1]*b[size-1];
}


void sparse_cov_mat_stat::fill_Csum(int which){
  int index,col,istart,istop,i,ii,isub;
  if(which==ALPHA || which==BETA){
    for(index=0;index<size;index++){
      diag_Csum[index]=data_diag_C[index]/divisor+(*sigma)*(*sigma)*diag_C[index];
    }
    for(index=0;index<maxlnz;index++){
      lnz_Csum[index]=data_lnz_C[index]/divisor+(*sigma)*(*sigma)*lnz_C[index];
    }
  }
  else{
    for(index=0;index<size;index++){
      diag_Csum[index]=pow(*delta+etahat[index]*data_diag_C[index],2.0)/(divisor*etahat[index]*etahat[index])+(*sigma)*(*sigma)*diag_C[index];
    }
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
        isub=nzsub[i-1]-1;
        // calculate vector of Euc distances between the pair of locations and store it
        lnz_Csum[index]=data_lnz_C[index]*data_lnz_C[index]/divisor+(*sigma)*(*sigma)*lnz_C[index];
        // delta is 0 off-diagonal so it plays no role here and things cancel out
        if((ii-1)!=index){
          cout << "Warning: ii and index not the same\n"; cout.flush();
        }
        i++; index++;
      }
    }
  }
}


void sparse_cov_mat_stat::gen_chol_Csum(void){
  // calculate numerical Cholesky factorization
  double eps=TOL;
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors

  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_Csum,temp_lnz_L);
  copy(size,diag_Csum,temp_diag_L);
  int warn=1;
  // Fortran routine for calculating numeric factorization
  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);

  if(warn!=1){
    cout << "Warning, zeroed-out column in Csum cholesky\n";
  }

  copy(maxlnz,temp_lnz_L,lnz_L_Csum);
  copy(size,temp_diag_L,diag_L_Csum);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}


int sparse_cov_mat_stat::solve_Csum(double* rhs_in, double* output){
  // calculates R^{-1} dot b where b=rhs_in
  // this follows the code of George & Liu's gsslv, p. 156-7

  // to solve Rb=L*t(L)b=y, let z=t(L)b, solve Lz=y for z; then solve
  // t(L)b=z for b;  i.e. y=R^{-1}b

  double rhsj, s;
  int i, ii, istop, istart,isub,j;

  // changed to double in regr4 for better avoiding underflow
  double* rhs=new double[size];
  double* diag=new double[size]; 
  double* lnz=new double[maxlnz];

  copy(maxlnz,lnz_L_Csum,lnz);  
  copy(size,diag_L_Csum,diag);  
  copy(size,rhs_in,rhs);
  
  // forward substitution -- solve Lz=y
  int flag=0;
  for(j=0;j<size;j++){
    if(diag[j]<TOL){
      rhs[j]=0.0;
      if(flag==0){
	flag=j;
      }
    }
    else{
      rhsj= rhs[j]/diag[j];
      rhs[j]= rhsj;
      //rhsj= (*rhs)(j)/diag[j];
      //(*rhs)(j)= rhsj;
      if(j==size-1 ) {break;}
      istart=xlnz[j];  
      if(j==size-2) {istop=istart;}
      else
	{
	  istop=xlnz[j+1]-1;
	}
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++)
	{
	  isub=nzsub[i-1];
	  rhs[isub-1]=rhs[isub-1]-lnz[ii-1]*rhsj;
	     //(*rhs)(isub-1)=(*rhs)(isub-1)-lnz[ii-1]*rhsj;
	  i++;
	}
    }
  }
  if(flag){
    cout << "Warning: zeroed out column in Csum matrix chol: " << flag << "\n"; cout.flush();
  }
  // backward substitution -- solve t(L)b=z


  for(j=size-1;j>=0;j--) {
    if(diag[j]<TOL){
      rhs[j]=0.0;
    }
    else{
      s=rhs[j];
      //s=(*rhs)(j);
      if(j<size-1){
	istart=xlnz[j];
	if(j==size-2) {istop=istart;}
	else{
	  istop=xlnz[j+1]-1;
	}
	i=xnzsub[j];
	for(ii=istart;ii<=istop;ii++){
	  isub=nzsub[i-1];
	  s=s-lnz[ii-1]*rhs[isub-1];
	  //s=s-lnz[ii-1]*(*rhs)(isub-1);
	  i++;
	}
      }
      rhs[j]=s/diag[j];
      //(*rhs)(j)=s/diag[j];
    }
  }
  
  copy(size,rhs,output);

  delete [] rhs;
  delete [] diag;
  delete [] lnz;
  return(flag);
}

void sparse_cov_mat::Ltmult(double* b,double* output){
  // calculates L^T*b

  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
      output[j]=0.0;
    }
  // j is col
  for(j=0;j<size-1;j++){
    if(diag_L[j]>=TOL){
      output[j]+=diag_L[j]*b[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	output[j]+=lnz_L[ii-1]*b[row];
	i++;
      }
    }
  }
  output[size-1]+=diag_L[size-1]*b[size-1];
}

void sparse_cov_mat_stat::Cparmmult(double* b,double* output){
  // calculates C*b, where C is C_{alpha,beta}

  int i, ii, istop, istart,j,row;

  for(j=0;j<size;j++){
    output[j]=0.0;
  }

  // j is col
  for(j=0;j<size-1;j++){
    output[j]+=diag_C[j]*b[j];

    istart=xlnz[j];
    istop=xlnz[j+1]-1;
    i=xnzsub[j];
    for(ii=istart;ii<=istop;ii++){
      row=nzsub[i-1]-1;
      output[row]+=lnz_C[ii-1]*b[j];
      output[j]+=lnz_C[ii-1]*b[row];
      i++;
    }
  }
  output[size-1]+=diag_C[size-1]*b[size-1];

  for(j=0;j<size;j++){
    output[j]*=(*sigma)*(*sigma);
  }
}

void sparse_cov_mat_stat::Cdatamult(double* b,double* output){
  // calculates C*b, where C is data covariance with divisor

  int i, ii, istop, istart,j,row;

  for(j=0;j<size;j++){
    output[j]=0.0;
  }

  // j is col
  for(j=0;j<size-1;j++){
    output[j]+=data_diag_C[j]*b[j];

    istart=xlnz[j];
    istop=xlnz[j+1]-1;
    i=xnzsub[j];
    for(ii=istart;ii<=istop;ii++){
      row=nzsub[i-1]-1;
      output[row]+=data_lnz_C[ii-1]*b[j];
      output[j]+=data_lnz_C[ii-1]*b[row];
      i++;
    }
  }
  output[size-1]+=data_diag_C[size-1]*b[size-1];

  for(j=0;j<size;j++){
    output[j]/=divisor;
  }
}

void sparse_cov_mat_stat::Cdata2mult(double* b,double* output){
  // calculates C*b, where C is cov of log(etahat)

  int i, ii, istop, istart,j,row;

  for(j=0;j<size;j++){
    output[j]=0.0;
  }

  // j is col
  for(j=0;j<size-1;j++){
    output[j]+=b[j]*pow(*delta+etahat[j]*data_diag_C[j],2.0)/(divisor*etahat[j]*etahat[j]);

    istart=xlnz[j];
    istop=xlnz[j+1]-1;
    i=xnzsub[j];
    for(ii=istart;ii<=istop;ii++){
      row=nzsub[i-1]-1;
      //      output[row]+=data_lnz_C[ii-1]*b[j];
      output[row]+=b[j]*pow(*delta+sqrt(etahat[j]*etahat[row])*data_lnz_C[ii-1],2.0)/(divisor*etahat[j]*etahat[row]);
      //output[j]+=data_lnz_C[ii-1]*b[row];
      output[j]+=b[row]*pow(*delta+sqrt(etahat[j]*etahat[row])*data_lnz_C[ii-1],2.0)/(divisor*etahat[j]*etahat[row]);
      i++;
    }
  }
  output[size-1]+=pow(*delta+etahat[size-1]*data_diag_C[size-1],2.0)/(divisor*etahat[size-1]*etahat[size-1])*b[size-1];
}



