/* File:      projection.C
   Date:      01/03/02  (written fall 2001)
   Author:    C. Paciorek - most functions taken from GMT source code
   Purpose:   independent functions for doing cartographic and projection calculations
   Compile:   g++ -c -L$(HOME)/lib/Proj/lib -lproj
   Usage:     
   Arguments: 
   Requires:  base.h, functions.h, projects.h (header file for Proj.4 library)
   Reference: most functions taken from GMT source code (mainly project.c): www.gmt.soest.hawaii.edu;
              project function relies on Proj.4 projection library: www.remotesensing.org/proj
*/

#include "base.h"
#include "projects.h"

void sincos (double a, double *s, double *c)
{
  *s = sin (a);
  *c = cos (a);
}

double GMT_mag3v (double *a)
{
  // computes ||a||

  return (sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]));
}


double GMT_dot3v (double *a, double *b)
{
  // computes a dot b

  return (a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);
}

void GMT_normalize3v (double *a)
{
  // normalizes a to length 1

  double r_length;
  r_length = (GMT_mag3v(a));
  if (r_length != 0.0) {
    r_length = 1.0 / r_length;
    a[0] *= r_length;
    a[1] *= r_length;
    a[2] *= r_length;
  }
}

void GMT_cross3v (double *a, double *b, double *c)
{
  // computes a X b

  c[0] = a[1] * b[2] - a[2] * b[1];
  c[1] = a[2] * b[0] - a[0] * b[2];
  c[2] = a[0] * b[1] - a[1] * b[0];
}

void GMT_geo_to_cart (double *alat, double *alon, double *a, int rads)
{
  /* Convert geographic latitude and longitude (alat, alon)
     to a 3-vector of unit length (a).  rads = TRUE if we
     need to convert alat, alon from degrees to radians  */
  
  double clat, clon, slon;
  
  if (rads) {
    *alat *= D2R;
    *alon *= D2R;
  }
  sincos (*alat, &a[2], &clat);
  sincos (*alon, &slon, &clon);
  a[0] = clat * clon;
  a[1] = clat * slon;
}

void GMT_cart_to_geo (double *alat, double *alon, double *a, int rads)
{
  /* Convert a 3-vector (a) of unit length into geographic
     coordinates (alat, alon).  rads = TRUE if we want the
     lat and lon converted from radians into degrees.  */
  
  if(rads) {
    *alat = R2D * asin(a[2]);
    *alon = R2D * atan2(a[1], a[0]);
  }
  else {
    *alat = asin(a[2]);
    *alon = atan2(a[1],a[0]);
  }
}

void	make_euler_matrix(double *p, double *e, double *theta, int rads)
{
  /* Routine to fill an euler matrix e with the elements
     needed to rotate a 3-vector about the pole p through
     an angle theta.  p is a unit 3-vector.  If rads = TRUE,
     we have to convert theta from degrees into radians before
     we use it.  */
  
  double	cos_theta, sin_theta, one_minus_cos_theta;
  double	pxsin, pysin, pzsin, temp;
  
  if (rads) {
    *theta *= D2R;
  }
  cos_theta = cos(*theta);
  sin_theta = sin(*theta);
  one_minus_cos_theta = 1.0 - cos_theta;
  
  pxsin = p[0] * sin_theta;
  pysin = p[1] * sin_theta;
  pzsin = p[2] * sin_theta;
  
  temp = p[0] * one_minus_cos_theta;
  e[0] = temp * p[0] + cos_theta;
  e[1] = temp * p[1] - pzsin;
  e[2] = temp * p[2] + pysin;
  
  temp = p[1] * one_minus_cos_theta;
  e[3] = temp * p[0] + pzsin;
  e[4] = temp * p[1] + cos_theta;
  e[5] = temp * p[2] - pxsin;
  
  temp = p[2] * one_minus_cos_theta;
  e[6] = temp * p[0] - pysin;
  e[7] = temp * p[1] + pxsin;
  e[8] = temp * p[2] + cos_theta;
}


void	matrix_3v(double *a, double *x, double *b)
{
  /* routine to find b, where Ax = b, A is a 3 by 3 square matrix,
     and x and b are 3-vectors.  A is stored row wise, that is:
     
     A = { a11, a12, a13, a21, a22, a23, a31, a32, a33 }  */
  
  b[0] = x[0]*a[0] + x[1]*a[1] + x[2]*a[2];
  b[1] = x[0]*a[3] + x[1]*a[4] + x[2]*a[5];
  b[2] = x[0]*a[6] + x[1]*a[7] + x[2]*a[8];
}


void	sphere_project_setup(double alat, double alon, double *a, double blat, double blon, double *b, double *azim, double *p, double *c, int two_pts, int rads)
{
  /* routine to initialize a pole vector, p, and a central meridian 
     normal vector, c, for use in projecting points onto a great circle.
     
     The great circle is specified in either one of two ways:
     if (two_pts), then the user has given two points, a and b,
     which specify the great circle (directed from a to b);
     if !(two_pts), then the user has given one point, a, and an azimuth,
     azim, clockwise from north, which defines the projection.
     
     The strategy is to use the oblique_transform operations above,
     in such a way that the great circle of the projection is the
     equator of an oblique transform, and the central meridian goes
     through a.  Then the transformed longitude gives the distance
     along the projection circle, and the transformed latitude gives
     the distance normal to the projection circle.
     
     If (two_pts), then p = normalized(a X b).  If not, we temporarily
     create p_temp = normalized(a X n), where n is the north pole.
     p_temp is then rotated about a through the angle azim to give p.
     After p is found, then c = normalized(p X a).
  */
  
  double	e[9];	/* Euler roatation matrix, if needed  */
  double neg_azim;
  
  /* First find p vector  */
  
  if (two_pts) {
    GMT_geo_to_cart(&alat, &alon, a, rads);
    GMT_geo_to_cart(&blat, &blon, b, rads);
    GMT_cross3v(a, b, p);
    GMT_normalize3v(p);
  }
  else {
    GMT_geo_to_cart(&alat, &alon, a, rads);
    b[0] = b[1] = 0.0;	/* set b to north pole  */
    b[2] = 1.0;
    GMT_cross3v(a, b, c);	/* use c for p_temp  */
    GMT_normalize3v(c);
    /*		make_euler_matrix(a, e, azim, rads);	*/
    neg_azim = -(*azim);
    make_euler_matrix(a, e, &neg_azim, rads);
    if (rads) *azim *= D2R;
    matrix_3v(e, c, p);	/* c (p_temp) rotates to p  */
  }
  
  /* Now set c vector  */
  
  GMT_cross3v(p, a, c);
  GMT_normalize3v(c);
}


void midpoint(double a_y, double a_x, double b_y, double b_x,double *mid_lat,double *mid_lon)
{
  /* routine to find great circle midpoint in geographic coordinates from two points,
     also in geographic coordinates */

  double e[9],a[3],b[3],pole[3],center[3],xt[3],x[3];
  int rads=TRUE;
  int two_points=TRUE;
  double azimuth;
  double l_max,d_to_km,xx,yy,theta;

  //d_to_km = 0.001 * 2.0 * PI * gmtdefs.ellipse[N_ELLIPSOIDS-1].eq_radius / 360.0;
  d_to_km = 0.001 * 2.0 * PI * EQ_RADIUS / 360.0;  // CJP changed to EQ_RADIUS 9/9/01
  sphere_project_setup(a_y, a_x, a, b_y, b_x, b, &azimuth, pole, center, two_points, rads);
  l_max = GMT_dot3v(a,b);
  l_max = acos(l_max) * R2D;   // gives distance from a to b

  /* this should give the same result as the formula for angular distance:
     diff=lon[i]-lon[j];
     if(diff<0) {diff=-diff;}
     dist=acos(sin(PI*lat[i]/180)*sin(PI*lat[j]/180)+cos(PI*lat[i]/180)*cos(PI*lat[j]/180)*cos(PI*diff/180));
     -- and it does on an example - CJP
  */

  xx = a_x;
  yy = a_y;
  GMT_geo_to_cart(&yy, &xx, x, rads);               // finds a in cart coords -> x
  theta = (double) l_max/2;                         // this computes half the angle
  make_euler_matrix(pole, e, &theta, rads);         // makes rotation matrix e
  matrix_3v(e,x,xt);                                // finds E*x=xt; i.e. rotates x by theta
  GMT_cart_to_geo(&yy, &xx, xt, rads);              // finds yy,xx in geo coords from xt
  *mid_lat = yy;
  *mid_lon = xx;       
  if(*mid_lon>=360.0) {*mid_lon-=360.0;}            // make sure between 0 and 360 ...
  if(*mid_lon<0.0) {*mid_lon+=360.0;}
}

void project(double baselat_val, double baselon_val, double xlat,double xlon,double ylat, double ylon, double *u,double *v,int units) {
  /* finds u and v distance between x and y where baselat and baselon are the geographic coordinates
     of the midpoint of x and y as calculated in the midpoint function */

  char* parms[4];
  parms[0]="proj=aeqd";           /* Azimuthal equidistance projection */
  parms[1]="ellps=clrk66";        /* model for the earth */

  string baselat="lat_0=";
  string baselon="lon_0=";
  
  baselat+=dtos(baselat_val);
  baselon+=dtos(baselon_val);
  baselat+="N";
  baselon+="E";
  parms[2]=(char*) baselat.c_str();
  parms[3]=(char*) baselon.c_str();

  PJ *ref;
  UV data;
  
  if(!(ref=pj_init(sizeof(parms)/sizeof(char *),parms)))
    {
      cerr << "Project initialization failed\n";
    }
  data.u=xlon*D2R;
  data.v=xlat*D2R;
  data=pj_fwd(data,ref);
  *u=data.u;
  *v=data.v;

  data.u=ylon*D2R;
  data.v=ylat*D2R;
  data=pj_fwd(data,ref);
  *u-=data.u;                      // *u is now Euclidean u distance between x and y locations
  *v-=data.v;                      // *v is now Euclidean v distance between x and y locations
  switch (units)
    {
    case MM:
      *u*=M2MM;
      *v*=M2MM;
      break;
    case KM:
      *u*=M2KM; 
      *v*=M2KM;
      break;
    default:
      cerr << "Cannot calculate Euclidean distance in units: " << units << endl;
      break;
    }
}
      






















